from collections import defaultdict
from dataclasses import dataclass
import math
import time
import types
import inspect
from datetime import timedelta
from threading import Event, Lock, Thread
from typing import (
    Any,
    Callable,
    DefaultDict,
    Hashable,
    Optional,
    Dict,
    Type,
    cast,
    Generic,
    TypeVar,
    overload,
)
from typing_extensions import Protocol, ParamSpec, Self, Concatenate
from theine_core import LruCore, TlfuCore
from theine.models import CachedValue
from functools import update_wrapper, _make_key
from uuid import uuid4

sentinel = object()


class Core(Protocol):
    def __init__(self, size: int):
        ...

    def schedule(self, key: str, expire: int):
        ...

    def deschedule(self, key: str):
        ...

    def set_policy(self, key: str) -> Optional[str]:
        ...

    def set(self, key: str, expire: int) -> Optional[str]:
        ...

    def remove(self, key: str):
        ...

    def access(self, key: str):
        ...

    def advance(self, now: int, cache: dict):
        ...


CORES: Dict[str, Type[Core]] = {
    "tlfu": TlfuCore,
    "lru": LruCore,
}

P = ParamSpec("P")
R = TypeVar("R")
_unset = object()


@dataclass
class EventData:
    event: Event
    data: Any


# https://github.com/python/cpython/issues/90780
class CachedAwaitable:
    def __init__(self, awaitable):
        self.awaitable = awaitable
        self.result = _unset

    def __await__(self):
        if self.result is _unset:
            self.result = yield from self.awaitable.__await__()
        return self.result


class Key:
    def __init__(self):
        self.key: Optional[str] = None
        self.event = Event()


class Wrapper(Generic[P, R]):
    def __init__(
        self,
        fn: Callable[P, R],
        timeout: Optional[timedelta],
        cache: "Cache",
        coro: bool,
        typed: bool,
        lock: bool,
    ):
        self._key_func: Optional[Callable[..., str]] = None
        self._hk_map: Dict[Hashable, str] = {}
        self._kh_map: Dict[str, Hashable] = {}
        self._events: Dict[Hashable, EventData] = {}
        self._func: Callable = fn
        self._cache: "Cache" = cache
        self._cache._enable_maintenance = False
        self._coro: bool = coro
        self._timeout: Optional[timedelta] = timeout
        self._typed: bool = typed
        self._auto_key: bool = True
        self._lock = lock
        update_wrapper(self, fn)

    def key(self, fn: Callable[P, str]) -> "Wrapper":
        self._key_func = fn
        self._cache._enable_maintenance = True
        self._auto_key = False
        return self

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        if self._auto_key:
            keyh = _make_key(args, kwargs, self._typed)
            v = self._hk_map.get(keyh, "")
            if v != "":
                key = v
            else:
                if self._lock:
                    event = EventData(Event(), None)
                    ve = self._events.setdefault(keyh, event)
                    if ve is event:
                        uid = key = uuid4().hex
                        self._hk_map[keyh] = uid
                        self._kh_map[uid] = keyh
                        event.data = uid
                        event.event.set()
                        self._events.pop(keyh, None)
                    else:
                        ve.event.wait()
                        key = cast(str, ve.data)
                else:
                    uid = key = uuid4().hex
                    self._hk_map[keyh] = uid
                    self._kh_map[uid] = keyh
        else:
            key = self._key_func(*args, **kwargs)  # type: ignore

        if self._coro:
            result = self._cache.get(key, sentinel)
            if result is sentinel:
                result = CachedAwaitable(self._func(*args, **kwargs))
                evicted = self._cache.set(key, result, self._timeout)
                if self._auto_key and evicted:
                    keyh = self._kh_map.pop(evicted, None)
                    if keyh:
                        self._hk_map.pop(keyh, None)
            return cast(R, result)

        data = self._cache.get(key, sentinel)
        if data is not sentinel:
            return cast(R, data)
        if self._lock:
            event = EventData(Event(), None)
            ve = self._events.setdefault(key, event)
            if ve is event:
                result = self._func(*args, **kwargs)
                event.data = result
                self._events.pop(key, None)
                evicted = self._cache.set(key, result, self._timeout)
                if self._auto_key and evicted:
                    keyh = self._kh_map.pop(evicted, None)
                    if keyh:
                        self._hk_map.pop(keyh, None)
                event.event.set()
            else:
                ve.event.wait()
                result = ve.data
        else:
            result = self._func(*args, **kwargs)
            evicted = self._cache.set(key, result, self._timeout)
            if self._auto_key and evicted:
                keyh = self._kh_map.pop(evicted, None)
                if keyh:
                    self._hk_map.pop(keyh, None)
        return cast(R, result)

    @overload
    def __get__(self, instance, owner) -> Callable[..., R]:
        ...

    @overload
    def __get__(self, instance, owner) -> Self:  # type: ignore
        ...

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return cast(Callable[..., R], types.MethodType(self, instance))


class Memoize:
    def __init__(
        self,
        cache: "Cache",
        timeout: Optional[timedelta],
        typed: bool = False,
        lock: bool = False,
    ):
        self.cache = cache
        self.timeout = timeout
        self.typed = typed
        self.lock = lock

    def __call__(self, fn: Callable[P, R]) -> Wrapper[P, R]:
        coro = inspect.iscoroutinefunction(fn)
        return Wrapper(fn, self.timeout, self.cache, coro, self.typed, self.lock)


class Cache:
    def __init__(self, policy: str, size: int):
        self._cache: Dict[Hashable, CachedValue] = {}
        self.core = CORES[policy](size)
        self._enable_maintenance = True
        self._maintainer = Thread(target=self.maintenance, daemon=True)
        self._maintainer.start()

    def __len__(self) -> int:
        return len(self._cache)

    def get(self, key: str, default: Any = None) -> Any:
        self.core.access(key)
        cached = self._cache.get(key, sentinel)
        if cached is sentinel:
            return default
        elif cast(CachedValue, cached).expire < time.time():
            self.delete(key)
            return default
        return cast(CachedValue, cached).data

    def set(
        self, key: str, value: Any, ttl: Optional[timedelta] = None
    ) -> Optional[str]:
        now = time.time()
        ts = max(ttl.total_seconds(), 1.0) if ttl is not None else math.inf
        expire = now + ts
        exist = key in self._cache
        v = CachedValue(value, expire)
        self._cache[key] = v
        if expire != math.inf:
            self.core.schedule(key, int(expire * 1e9))
        else:
            self.core.deschedule(key)
        if exist:
            return None
        evicted = self.core.set_policy(key)
        if evicted is not None:
            self._cache.pop(evicted, None)
            return evicted
        return None

    def delete(self, key: str) -> bool:
        v = self._cache.pop(key, sentinel)
        if v is not sentinel:
            self.core.remove(key)
            return True
        return False

    def maintenance(self):
        while True:
            if self._enable_maintenance:
                self.core.advance(time.time_ns(), self._cache)
            time.sleep(0.5)



from threading import Lock
import time
from datetime import timedelta
from typing import Optional, cast
from theine import Cache as Theine
from django.core.cache.backends.base import DEFAULT_TIMEOUT, BaseCache


class Cache(BaseCache):
    def __init__(self, name, params):
        super().__init__(params)
        self.cache = Theine("tlfu", self._max_entries)

    def _timeout_seconds(self, timeout) -> Optional[float]:
        if timeout == DEFAULT_TIMEOUT:
            return self.default_timeout
        return timeout

    def add(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_key(key, version)
        if key in self.cache._cache and self.cache._cache[key].expire > time.time():
            return False
        self.cache.set(
            key,
            value,
            timedelta(seconds=cast(float, timeout))
            if timeout is not DEFAULT_TIMEOUT
            else None,
        )
        return True

    def get(self, key, default=None, version=None):
        key = self.make_key(key, version)
        return self.cache.get(key, default)

    def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        to = self._timeout_seconds(timeout)
        if to is not None and to <= 0:
            self.delete(key)
            return
        key = self.make_key(key, version)
        print("setttt", key, to)
        self.cache.set(
            key,
            value,
            timedelta(seconds=to) if to is not None else None,
        )

    def touch(self, key, timeout=DEFAULT_TIMEOUT, version=None):
        nkey = self.make_key(key, version)
        if nkey in self.cache._cache:
            with Lock():
                self.set(key, self.cache._cache[nkey].data, timeout, version)
            return True
        return False

    def delete(self, key, version=None):
        key = self.make_key(key, version)
        return self.cache.delete(key)

    def clear(self):
        for key in self.cache._cache:
            self.cache.core.remove(cast(str, key))
        self.cache._cache = {}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['theine', 'theine.adapters']

package_data = \
{'': ['*']}

install_requires = \
['theine-core>=0.1.6,<0.2.0', 'typing-extensions>=4.4.0,<5.0.0']

setup_kwargs = {
    'name': 'theine',
    'version': '0.1.3',
    'description': 'high performance in-memory cache',
    'long_description': '# Theine\nHigh performance in-memory cache inspired by [Caffeine](https://github.com/ben-manes/caffeine).\n\n- High performance [Rust core](https://github.com/Yiling-J/theine-core)\n- High hit ratio with [W-TinyLFU eviction policy](https://arxiv.org/pdf/1512.00727.pdf)\n- Expired data are removed automatically using [hierarchical timer wheel](http://www.cs.columbia.edu/~nahum/w6998/papers/ton97-timing-wheels.pdf)\n- Simple API\n- Django cache backend\n\n## Table of Contents\n\n- [Requirements](#requirements)\n- [Installation](#installation)\n- [API](#api)\n- [Decorator](#decorator)\n- [Django Cache Backend](#django-cache-backend)\n- [Benchmarks(WIP)](#benchmarks)\n  * [continuous benchmark](#continuous-benchmark)\n  * [10k requests zipf](#10k-requests-zipf)\n\n## Requirements\nPython 3.7+\n\n## Installation\n```\npip install theine\n```\n\n## API\n\n```Python\nfrom theine import Cache\nfrom datetime import timedelta\n\ncache = Cache("tlfu", 10000)\n# without default, return None on miss\nv = cache.get("key")\n\n# with default, return default on miss\nsentinel = object()\nv = cache.get("key", sentinel)\n\n# set with ttl\ncache.set("key", {"foo": "bar"}, timedelta(seconds=100))\n\n# delete from cache\ncache.delete("key")\n```\n\n## Decorator\nTheine support string keys only, so to use a decorator, a function to convert input signatures to string is necessary. **The recommended way is specifying the function explicitly**, this is approach 1, Theine also support generating key automatically, this is approach 2. I will list pros and cons below.\n\n**- explicit key function**\n\n```python\nfrom theine import Cache, Memoize\nfrom datetime import timedelta\n\n@Memoize(Cache("tlfu", 10000), timedelta(seconds=100))\ndef foo(a:int) -> int:\n    return a\n\n@foo.key\ndef _(a:int) -> str:\n    return f"a:{a}"\n\nfoo(1)\n\n# asyncio\n@Memoize(Cache("tlfu", 10000), timedelta(seconds=100))\nasync def foo_a(a:int) -> int:\n    return a\n\n@foo_a.key\ndef _(a:int) -> str:\n    return f"a:{a}"\n\nawait foo_a(1)\n\n```\n\n**Pros**\n- A decorator with both sync and async support, you can replace your lru_cache with Theine now.\n- Thundering herd protection(multithreading: set `lock=True` in `Memoize`, asyncio: always enabled).\n- Type checked. Mypy can check key function to make sure it has same input signature as original function and return a string.\n\n**Cons**\n- You have to use 2 functions.\n- Performance. Theine API: around 8ms/10k requests ->> decorator: around 12ms/10k requests.\n\n**- auto key function**\n\n```python\nfrom theine import Cache, Memoize\nfrom datetime import timedelta\n\n@Memoize(Cache("tlfu", 10000), timedelta(seconds=100), typed=True)\ndef foo(a:int) -> int:\n    return a\n\nfoo(1)\n\n# asyncio\n@Memoize(Cache("tlfu", 10000), timedelta(seconds=100), typed=True)\nasync def foo_a(a:int) -> int:\n    return a\n\nawait foo_a(1)\n\n```\n**Pros**\n- Same as explicit key version.\n- No extra key function.\n\n**Cons**\n- Worse performance: around 18ms/10k requests.\n- Auto removal of stale keys is disabled due to current implementation.\n- Unexpected memory usage. The auto key function use same methods as Python\'s lru_cache. Take a look [this issue](https://github.com/python/cpython/issues/88476) or [this one](https://github.com/python/cpython/issues/64058).\n\n\n## Django Cache Backend\n\n```Python\nCACHES = {\n    "default": {\n        "BACKEND": "theine.adapters.django.Cache",\n        "TIMEOUT": 300,\n        "OPTIONS": {"MAX_ENTRIES": 10000},\n    },\n}\n```\n\n## Benchmarks\n### continuous benchmark\nhttps://github.com/Yiling-J/cacheme-benchmark\n\n### 10k requests zipf\nCachetools: https://github.com/tkem/cachetools\n\nSource Code: https://github.com/Yiling-J/theine/blob/main/benchmarks/benchmark_test.py\n\n|       | Theine API | Theine(W-TinyLFU) Custom-Key Decorator | Theine(W-TinyLFU) Auto-Key Decorator | Cachetools(LFU) Decorator |\n|-------|------------|----------------------------------------|--------------------------------------|---------------------------|\n| Read  | 6.03 ms    | 10.75 ms                               | 12.75 ms                             | 17.10 ms                  |\n| Write | 23.22 ms   | 26.22 ms                               | 67.53 ms                             | 440.50 ms                 |\n\n',
    'author': 'Yiling-J',
    'author_email': 'njjyl723@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
